#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "grb2.h"
#include "wgrib2.h"
#include "fnlist.h"

/*
 * Fix_ncep.c
 *
 * 7/2008: Public Domain: Wesley Ebisuzaki
 *
 */

extern int decode, latlon;
extern double *lat, *lon;
extern int nx, ny;

/*
 * HEADER:100:fix_ncep:misc:0:fix ncep PDT=8 headers produced by cnvgrib2
 */

int f_fix_ncep(ARG0) {
    unsigned char *p;
    int unit, n, type;
    unsigned int u, NAve;
    int year, month, day, hour, minute, second;
    int units;
    unsigned int dtime;

    if (mode < 0) return 0;

    if (GB2_Center(sec) != NCEP) return 0;		// not NCEP

    if (GB2_ProdDefTemplateNo(sec) != 8) return 0;

    p = sec[4];

    if ((unit = code_table_4_4(sec)) < 0) return -1;
    n = (int) p[41];
    if (mode == 99) fprintf(stdout,"\nn=sec[4][41]=%d\n", n);
    type = p[46];
    if (mode == 99) fprintf(stdout,"type=sec[4][46]=%d\n", type);

    if (mode == 99) {
        wrt_time(p[17], uint4(p+18),inv_out);
	fprintf(stdout,"P1=%s  ", inv_out);
	fprintf(stdout,"Nave=%d  type=%d\n", uint4(p+42), type);
        wrt_time(p[48], uint4(p+49),inv_out);
	fprintf(stdout,"time1=%s", inv_out);

        wrt_time(p[53], uint4(p+54),inv_out);
	fprintf(stdout," time2=%s\n",inv_out);

	if (n >= 2) {
        wrt_time(p[48+12], uint4(p+49+12),inv_out);
	fprintf(stdout,"time1=%s", inv_out);
        wrt_time(p[53+12], uint4(p+54+12),inv_out);
        }
	fprintf(stdout," time2=%s\n",inv_out);
	fprintf(stdout," time: %d-%d-%d-%d\n", int2(p+34), p[36], p[37], p[38]);
	fprintf(stdout," Nave=%d\n",uint4(p+49));
    }

    NAve = uint4(p+49);

//  fix up end of interval time

    if (type == 193 || type == 194 || type == 195 || type == 196 || type == 197
	|| type == 203 || type == 204 || type == 205 || type == 207) {

	// get reference time
        get_time(sec[1]+12, &year, &month, &day, &hour, &minute, &second);

	// add in forecast time

        units = sec[4][17];
        dtime = uint4(sec[4]+18);
        add_time(&year, &month, &day, &hour, &minute, &second, dtime, units);

	// add (Nave -1)*dt
	if (type == 194 || type == 196) {
	    units = 2;			// dt = 1 day
	    dtime = 1;
	}
	else if (type == 203 || type == 204) {
	    units = 1;			// dt = 6 hours
	    dtime = 6;
	}
	else if (type == 205 || type == 206) {
	    units = 1;			// dt = 12 hours
	    dtime = 12;
	}
	else {
            units = p[53];
            dtime = uint4(p+54);
	}
        dtime *= (uint4(p+49)-1);
        add_time(&year, &month, &day, &hour, &minute, &second, dtime, units);

	// if n == 2, add to time
	if (sec[4][41] == 2) {
	    units = sec[4][60];
	    dtime = uint4(p+61);
            add_time(&year, &month, &day, &hour, &minute, &second, dtime, units);
	}
	if (sec[4][41] > 2) {
	    fprintf(stderr,"Help fix_ncep n = %d\n", (int) sec[4][41] );
	}

        save_time(year,month,day,hour,minute,second, sec[4]+34);
    }

    
    if (type == 192 && n == 2) {		// climo

        // get reference time
        get_time(sec[1]+12, &year, &month, &day, &hour, &minute, &second);

        // add in forecast time

        units = sec[4][17];
        dtime = uint4(sec[4]+18);
        add_time(&year, &month, &day, &hour, &minute, &second, dtime, units);

        if (sec[4][41] == 2) {
            units = sec[4][60];
            dtime = uint4(p+61);
            add_time(&year, &month, &day, &hour, &minute, &second, dtime, units);
        }
        if (sec[4][41] > 2) {
            fprintf(stderr,"Help fix_ncep n = %d\n", (int) sec[4][41] );
        }

        save_time(year,month,day,hour,minute,second, sec[4]+34);



	p[46] = 51;
	p[47] = 1;

	p[48] = 4;		// outer loop NAve years long
	u = NAve;
	uint_char(u,p+49);

	p[53] = 4;		// 1 year between fields
	u = 1;
	uint_char(u,p+54);

    }

    if (type == 193) {
	p[46] = 0;		// average
	p[47] = 1;
	u = uint4(p+49) * uint4(p+54);
	uint_char(u, p+49);
    }

    if (type == 194) {		// ref cnvgrib
        p[46] = 0;	// average
	p[48] = 2;	// units=day
	p[53] = 2;	// units=day
	u = 1;
	uint_char(u, p+54);
        p[58] = 1;	// accumulation
    }

    if (type == 195) {
	p[58] =  1; // accumulation
	p[59] = 2;

	p[60] = p[53];
	p[61] = p[54];
	p[62] = p[55];
	p[63] = p[56];
	p[64] = p[57];

	p[46] = 0;
	u = uint4(p+49) * uint4(p+54);
	uint_char(u, p+49);
    }

    if (type == 196) {	// cnvgrib
        p[46] = 0;	// average
	p[48] = 2;	// day
	p[53] = 1;	// hour
        p[58] = 0;	// average
    }

    if (type == 197) {
        p[58] =  0; // average
        p[59] = 2;

        p[60] = p[53];
        p[61] = p[54];
        p[62] = p[55];
        p[63] = p[56];
        p[64] = p[57];

        p[46] = 1;		// accumulation
        p[47] = 2;		// ft+
        u = uint4(p+49) * uint4(p+54);
        uint_char(u, p+49);

    }


    return 0;
}

