#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "grb2.h"
#include "wgrib2.h"
#include "fnlist.h"

/*
 * RPN reversed polish notation
 *
 * 4/2009 Public Domain by Wesley Ebisuzaki
 *
 * operations:
 *
 *     + - * /
 *    == != < <= > >=
 *
 *    sqrt 1/x floor
 *    min, max, merge
 *
 *    pi
 *
 *    registers: sto_N, rcl_N, clr_N
 *               rcl_lat, rcl_lon
 *               rcl (data)
 *    stack: exc, pop, dup, clr
 *
 * at the end of rpn, the top of the stack is saved to data unless clr done first
 */
#define N_REGS 10
#define STACK_SIZE 10

extern int decode, latlon;
extern double *lat, *lon;
extern int match_flag;

int rpn_n[N_REGS];
float *rpn_data[N_REGS];
static float *stack[STACK_SIZE];

#define SCALAR 0
#define VECTOR 1
#define DBL_VEC 2

int push(int top, int ndata, int type, float f, float *ff, double *d);


/*
 * HEADER:100:rpn:misc:1:reverse polish notation calculator (beta)
 */

int f_rpn(ARG1) {
    char string[100], *p;
    int i, j, n;
    float f;
    int top;

    static int state=0;

    if (mode == -1) {
	decode = latlon = 1;
	if (state == 0) {
            for (i = 0; i < N_REGS; i++) {
	        rpn_n[i] = 0;
	        rpn_data[i] = NULL;
	    }
	    state = 1;
	}
	return 0;
    }
    if (mode == -2) {
	if (state == 1) {
            for (i = 0; i < N_REGS; i++) {
	        if (rpn_data[i]) {
		    free (rpn_data[i]);
		    rpn_data[i] = NULL;
	            rpn_n[i] = 0;
	        }
	    }
	}
	state = 0;
	return 0;
    }


    // initialize stack

    if (data == NULL) fatal_error("-rpn: decode failed","");

    for (i = 0; i < STACK_SIZE; i++) stack[i] = NULL;
    top = push(-1, ndata, VECTOR, 0.0, data, NULL);

    // scan parameters
   
    p = arg1;
    while (sscanf(p,"%[^:]%n", string, &n) == 1) {
	if (mode == 99) { 
	    sprintf(inv_out,"string=(%s) top=%d", string, top);
	    printf("string=(%s) top=%d", string, top);
	    inv_out += strlen(inv_out);
	}
	p = p + n;
	if (*p == ':') p++;

	// binary operators + - * /

	if (strcmp(string,"+") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad + expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    stack[j][i] = stack[j][i] + stack[top][i];
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}


	else if (strcmp(string,"-") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad - expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    stack[j][i] = stack[j][i] - stack[top][i];
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}
	else if (strcmp(string,"*") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad * expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    stack[j][i] = stack[j][i] * stack[top][i];
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}
	else if (strcmp(string,"/") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad / expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i]) && (stack[top][i] != 0.0)) {
		    stack[j][i] = stack[j][i] / stack[top][i];
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}

	// merge:  stack(top-1) = stack(top) (if defined) ; top--;

	else if (strcmp(string,"merge") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad merge expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    stack[j][i] = stack[top][i];
		}
	    }
	    top--;
	}


	// exc (swap top and top-1 stack entries)

	else if (strcmp(string,"exc") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad exec expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		f = stack[j][i];
		stack[j][i] = stack[top][i];
		stack[top][i] = j;
	    }
	    top--;
	}

	// pop:  top--;

	else if (strcmp(string,"pop") == 0) {
	    if (top < 0) fatal_error("-rpn: bad pop","");
	    top--;
	}

	// dup: top++; stack(top) = stack(top-1)

	else if (strcmp(string,"dup") == 0) {
	    top = push(top,ndata,VECTOR,0.0,stack[top],NULL);
	}

	//  sqrt: stack(top) = sqrt(stack(top))

	else if (strcmp(string,"sqrt") == 0) {
	    if (top < 0) fatal_error("-rpn: bad sqrt expression","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && stack[top][i] >= 0.0) {
		    stack[top][i] = sqrt(stack[top][i]);
		}
		else stack[top][i] = UNDEFINED;
	    }
	}
	// sq: x*x
	else if (strcmp(string,"sq") == 0) {
	    if (top < 0) fatal_error("-rpn: bad sq expression","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    stack[top][i] *= stack[top][i];
		}
	    }
	}
	// abs
	else if (strcmp(string,"abs") == 0) {
	    if (top < 0) fatal_error("-rpn: bad abs expression","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    if (stack[top][i] < 0.0) stack[top][i] = -stack[top][i];
		}
	    }
	}


	// 1/x

	else if (strcmp(string,"1/x") == 0) {
	    if (top < 0) fatal_error("-rpn: bad 1/x","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && stack[top][i] != 0.0) {
		    stack[top][i] = 1.0 / stack[top][i];
		}
		else stack[top][i] = UNDEFINED;
	    }
	}

	// floor

	else if (strcmp(string,"floor") == 0) {
	    if (top < 0) fatal_error("-rpn: bad int","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    stack[top][i] = floor(stack[top][i]);
		}
	    }
	}

	// sin cos tan
	else if (strcmp(string,"sin") == 0) {
	    if (top < 0) fatal_error("-rpn: bad sin","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    stack[top][i] = sinf(stack[top][i]);
		}
	    }
	}
	else if (strcmp(string,"cos") == 0) {
	    if (top < 0) fatal_error("-rpn: bad cos","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    stack[top][i] = cosf(stack[top][i]);
		}
	    }
	}
	else if (strcmp(string,"tan") == 0) {
	    if (top < 0) fatal_error("-rpn: bad tan","");
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i])) {
		    stack[top][i] = tanf(stack[top][i]);
		}
	    }
	}

	// sto_N

	else if (string[0] == 's' && string[1] == 't' && string[2] == 'o' && string[3] == '_' 
		&& isdigit(string[4]) && string[5] == 0) {
	    if (top < 0) fatal_error("-rpn: sto","");
	    i = atoi(string+4);
	    if (i >= N_REGS || i < 0) fatal_error("rpn: register number in %s", string);
	    if (ndata != rpn_n[i]) {
		if (rpn_data[i]) free(rpn_data[i]);
		rpn_n[i] = ndata;
		rpn_data[i] = (float *) malloc(ndata* sizeof(float));
		if (rpn_data[i] == NULL) fatal_error("rpn: memory allocation","");
	    }
	    for (j=0; j < ndata; j++) {
		rpn_data[i][j] = stack[top][j];
	    }
        }

	// rcl_N
	else if (string[0] == 'r' && string[1] == 'c' && string[2] == 'l' && string[3] == '_'
		&& isdigit(string[4]) && string[5] == 0) {
	    i = atoi(string+4);
	    if (i >= N_REGS || i < 0) fatal_error("rpn: register number in %s", string);

	    if (rpn_n[i] != 0 && rpn_n[i] != ndata) fatal_error("rpn: rcl size mismatch","");

	    if (rpn_n[i] == 0) {	// unused register are zero
	        top = push(top,ndata,SCALAR,0.0,rpn_data[i],NULL);
	    }
	    else {
	        top = push(top,ndata,VECTOR,0.0,rpn_data[i],NULL);
	    }
	}

	// clr_N
	else if (string[0] == 'c' && string[1] == 'l' && string[2] == 'r' && string[3] == '_'
		&& isdigit(string[4]) && string[5] == 0) {
	    i = atoi(string+4);
	    if (i >= N_REGS || i < 0) fatal_error("rpn: register number in %s", string);
	    if (rpn_data[i]) {
		free(rpn_data[i]);
		rpn_data[i] = NULL;
	    }
	    rpn_n[i] = 0;
	}

	// rcl_lat
	else if (strcmp(string,"rcl_lat") == 0) {
	    if (lat == NULL) fatal_error("rpn: lat not defined","");
	    top = push(top,ndata,DBL_VEC,0.0,NULL,lat);
	}
	// rcl_lon
	else if (strcmp(string,"rcl_lon") == 0) {
	    if (lon == NULL) fatal_error("rpn: lon not defined","");
	    top = push(top,ndata,DBL_VEC,0.0,NULL,lon);
	}

	// max and min

	else if (strcmp(string,"max") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad max expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    if (stack[j][i] < stack[top][i]) stack[j][i] = stack[top][i];
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}
	else if (strcmp(string,"min") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad min expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    if (stack[j][i] > stack[top][i]) stack[j][i] = stack[top][i];
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}

	else if (strcmp(string,">") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad > expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    stack[j][i] =  (stack[j][i] > stack[top][i]);
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}
        else if (strcmp(string,">=") == 0) {
            if (top <= 0) fatal_error("-rpn: bad >= expression","");
            j = top-1;
            for (i = 0; i < ndata; i++) {
                if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
                    stack[j][i] =  (stack[j][i] >= stack[top][i]);
                }
                else stack[j][i] = UNDEFINED;
            }
            top--;
        }
        else if (strcmp(string,"!=") == 0) {
            if (top <= 0) fatal_error("-rpn: bad > expression","");
            j = top-1;
            for (i = 0; i < ndata; i++) {
                if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
                    stack[j][i] =  (stack[j][i] != stack[top][i]);
                }
                else stack[j][i] = UNDEFINED;
            }
            top--;
        }
	else if (strcmp(string,"==") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad == expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    stack[j][i] =  (stack[j][i] == stack[top][i]);
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}
        else if (strcmp(string,"<") == 0) {
            if (top <= 0) fatal_error("-rpn: bad < expression","");
            j = top-1;
            for (i = 0; i < ndata; i++) {
                if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
                    stack[j][i] =  (stack[j][i] < stack[top][i]);
                }
                else stack[j][i] = UNDEFINED;
            }
            top--;
        }
        else if (strcmp(string,"<=") == 0) {
            if (top <= 0) fatal_error("-rpn: bad <= expression","");
            j = top-1;
            for (i = 0; i < ndata; i++) {
                if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
                    stack[j][i] =  (stack[j][i] <= stack[top][i]);
                }
                else stack[j][i] = UNDEFINED;
            }
            top--;
        }

	else if (strcmp(string,"mask") == 0) {
	    if (top <= 0) fatal_error("-rpn: bad mask expression","");
	    j = top-1;
	    for (i = 0; i < ndata; i++) {
		if (DEFINED_VAL(stack[top][i]) && DEFINED_VAL(stack[j][i])) {
		    if (stack[top][i] == 0.0) stack[j][i] = UNDEFINED;
	        }
		else stack[j][i] = UNDEFINED;
	    }
	    top--;
	}

	// change to rcl-data, rcl-lat, rcl-lon

	// rcl:  stack(++top) = data
	else if (strcmp(string,"rcl") == 0) {
	    top = push(top,ndata,VECTOR,0.0,data,NULL);
	}

	// sto:  data = stack(top)
	else if (strcmp(string,"sto") == 0) {
	    if (top < 0) fatal_error("-rpn: sto","");
	    for (j=0; j < ndata; j++) {
		data[j] = stack[top][j];
	    }
        }

	// clr: emtpy stack
	else if (strcmp(string,"clr") == 0) {
	    top = -1;
	}

	// pi: stack(++top) = pi
	else if (strcmp(string,"pi") == 0) {
	    top = push(top,ndata,SCALAR,(float) M_PI,NULL,NULL);
	}

	// number:  stack(++top) = number

	else if (string[0] == '+' || string[0] == '-' || isdigit(string[0])) {
	    f = atof(string);
	    top = push(top,ndata,SCALAR,f,NULL,NULL);
	}
	else fatal_error("rpn: unidentified symbol %s", string);
    }	

    if (top >= 0) {
	for (i = 0; i < ndata; i++) {
	    data[i] = stack[top][i];
	}
    }
    else fatal_error("rpn: stack empty","");
    return 0;
}

int push(int top, int ndata, int type, float f, float *ff, double *d) {

    int i;

    if (++top == STACK_SIZE) fatal_error_i("rpn: stack overflow %d",top);
    if (stack[top] == NULL) {
	 stack[top] = (float *) malloc(ndata * sizeof(float));
	if (stack[top] == NULL) fatal_error("rpn: memory allocation","");
    }
    if (type == SCALAR) {
	for (i = 0; i < ndata; i++) stack[top][i] = f;
    }
    else if (type == VECTOR) {
	for (i = 0; i < ndata; i++) stack[top][i] = ff[i];
    }
    else if (type == DBL_VEC) {
	for (i = 0; i < ndata; i++) stack[top][i] = (float) d[i];
    }

    return top;
}

/*
 * HEADER:100:if_reg:misc:1:if rpn registers defined, X = A, A:B, A:B:C, etc A = register number
 */
int f_if_reg(ARG1) {
    int i, j, *list;
    char *p;

    if (mode == -1) {
	// figure out the number of arguments
	i = 1;
	p = arg1;
	while (*p) {
	   if (*p++ == ':') i++;
	}
	*local = list = (int *) calloc(i+1, sizeof (int));
	if (list == NULL) fatal_error("if_reg: memory allocation failed","");
	list[0] = i;
	p = arg1;
	for (j = 1; j <= i; j++) {
	    list[j] = atoi(p);
	    if (list[j] >= N_REGS || list[j] < 0) fatal_error_i("if_reg: bad register %d", list[j]);
	    while (isdigit(*p)) p++;
	    if (*p == ':') p++;
	}
    }
    else if (mode == -2) {
	list = (int *) *local;
	free(list);
    }
    else if (mode >= 0) {
	list = (int *) *local;
	i = list[0];
	for (j=1; j <= i; j++) {
	    if (rpn_n[list[j]] == 0) { match_flag = 1; return 0; }
	}
    }
    return 0;
}
